/*
** Module   : Function.java
** Abstract : Interface which defines basic, required function services
**
** Copyright (c) 2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**                   Golden Code Development Corporation
**                            CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- -----------------------------Description-----------------------------
** 001 ECF 20050215   @19821 Created initial version. Defines API for a user function to provide
**                           basic services.
** 002 ECF 20050222   @20073 Changed signature of invoke method to throw SymbolException. 
**                           FunctionException is now obsolete.
** 003 ECF 20050303   @20177 Added support for variable argument lists (varargs). Removed
**                           getSignature and added matchSignature method.
** 004 ECF 20050509   @21376 Complete rewrite. Previously an interface to be implemented for
**                           runtime method invocation support, this class is now used for parse
**                           time resolution of method references in user expressions.
** 005 ECF 20050607   @21437 Always store parameter types specified at construction. Previously,
**                           we only used these for methods which supported var-arg lists,
**                           but they are required when explicitly casting parameter types for
**                           regular methods as well.
** 006 ECF 20050608   @21454 Fix to isAssignmentCompatible method. Was reporting false positives
**                           in some cases when a provided primitive type was being compared
**                           to a required object reference type.
** 007 OM  20150122          Added boolean to the list of automatically boxing/unboxing types in
**                           strict mode matching.
*/

package com.goldencode.expr;

import java.lang.reflect.*;

/**
 * Support class which assists in parse-time resolution of method references
 * in user expressions. Manages data used by the {@link SymbolResolver} on
 * behalf of the {@link ExpressionParser} when introspecting classes to match
 * method invocation semantics in a user expression, with a target method
 * in a Java class. The following information is available:
 * <ul>
 * <li>an invocation target class;
 * <li>the return type which is expected by the caller of the underlying
 *     method (which may be different from the method's return type,
 *     requiring some conversion);
 * <li>the target method;
 * <li>the parameter types required by the function (which may be different
 *     than those dictated by the underlying method, in the case of a method
 *     which accepts a variable-length argument list);
 * <li>whether or not to simulate a variable-length argument list when
 *     invoking the underlying method, and if so, how many arguments are
 *     required, at minimum;
 * <li>the category of the underlying method:  general purpose, or one of the
 *     a special, <a href="#bootstrap">bootstrap</a> methods;
 * <li>the prerequisite function, if any, which must be invoked in order to
 *     retrieve the invocation target object for this function's invocation;
 * <li>an optional symbol which represents a name or identifier for the
 *     invocation target object (used to set the text of an AST node when
 *     depicting an expression in postfix form).
 * </ul>
 * <p>
 * This class provides methods to conduct fuzzy matches of method parameter
 * signatures to one another. Fuzzy matches are those which match parameters
 * of {@link #isAssignmentCompatible assignment compatible} types. These may
 * not be exact type matches as required by the VM when invoking methods,
 * but the assumption is that the {@link Compiler expression compiler} will
 * be able to generate the necessary bytecode at assemble time to convert
 * the provided parameter type to the required parameter type. This will
 * occur through the assembly of the necessary instructions to wrap/unwrap
 * primitive values, cast object references, and/or perform widening or
 * narrowing conversions on primitive values. The primary worker method which
 * performs this fuzzy matching is {@link #matchSignature(Class[], Class[],
 * int, boolean)}.
 * <p>
 * <a name="bootstrap"><b>Bootstrap Methods.</b> Several methods are treated
 * specially by the parser, symbol resolver, and compiler. All but the most
 * trivial user expressions must access application specific data. The access
 * points for this data are {@link SymbolResolver#registerLibrary callback
 * libraries} and {@link SymbolResolver#registerVariable user variables},
 * which the application registers with the expression engine. A compiled
 * expression must have a way to access these constructs. When a reference
 * to a callback library object or a user variable is encountered in an
 * expression, the compiler embeds instructions into the expression object
 * to invoke methods which access these objects at expression execution time.
 * These are termed bootstrap methods, because they permit access to these
 * primordial access points into the application's data. There are currently
 * three bootstrap methods supported:
 * <ul>
 * <li>{@link CompiledExpression#getLib} - loads a callback library object,
 *     so that the expression can invoke a method upon it.
 * <li>{@link CompiledExpression#getVar} - loads the <em>current value</em>
 *     of a user variable, so that the expression can invoke a method upon
 *     it, use it in an operation, or return it.
 * <li>{@link CompiledExpression#setVar} - sets the <em>current value</em>
 *     of a user variable as the result of an assignment operation in the
 *     expression.
 * </ul>
 * <p>
 * At the time an expression is parsed, it is determined whether a bootstrap
 * method invocation must be compiled into that expression. An instance of
 * this class is created accordingly, and its type is set to the appropriate
 * constant to represent which of the above bootstrap methods it represents.
 */
public final class Function
{
   /** General purpose method */
   static final int GENERAL = 0;
   
   /** Bootstrap method which retrieves a callback library object */
   static final int GET_LIB = 1;
   
   /** Bootstrap method which retrieves the value of a variable */
   static final int GET_VAR = 2;
   
   /** Bootstrap method which sets the value of a variable */
   static final int SET_VAR = 3;
   
   /** Pairs of data types used to test strict assignment compatibility */
   private static final Class[] COMPAT_PAIRS =
   {
      Integer.TYPE   ,  Integer.class   ,
      Long.TYPE      ,  Long.class      ,
      Double.TYPE    ,  Double.class    ,
      Float.TYPE     ,  Float.class     ,
      Short.TYPE     ,  Short.class     ,
      Character.TYPE ,  Character.class ,
      Byte.TYPE      ,  Byte.class      ,
      Boolean.TYPE   ,  Boolean.class   ,
   };
   
   /** Invocation target class of the compiled version of this function */
   private Class target = null;
   
   /** Type of object which must be returned by this function */
   private Class returnType = null;
   
   /** Method invoked by the compiled version of this function */
   private Method method = null;
   
   /** Parameter types required by the method */
   private Class[] parameterTypes = null;
   
   /** Number of required arguments in a var-arg scenario */
   private int minArgs = -1;
   
   /** Method type represented by this object */
   private int type = GENERAL;
   
   /** Symbol assigned to this object for use in AST building */
   private String symbol = null;
   
   /** Function which must be executed to provide this function's target */
   private Function prerequisite = null;
   
   /**
    * Construct an instance of this object to represent a regular method
    * which does not accept variable-length argument lists.
    *
    * @param   target
    *          Invocation target class.
    * @param   returnType
    *          Type of object which caller of this method expects to receive.
    * @param   method
    *          Target method to be invoked.
    */
   Function(Class target, Class returnType, Method method)
   {
      this(target, returnType, method, null, false);
   }
   
   /**
    * Construct an instance of this object to represent a method which may
    * or may not accept variable-length argument lists.
    *
    * @param   target
    *          Invocation target class.
    * @param   returnType
    *          Type of object which caller of this method expects to receive.
    * @param   method
    *          Target method to be invoked.
    * @param   parameterTypes
    *          Array of parameter types required by the underlying method.
    *          These may be different than the underlying method reports,
    *          in the case of methods which use var-args. Only used if the
    *          underlying method accepts a var-arg list; may be
    *          <code>null</code> otherwise.
    * @param   enableVarArgs
    *          <code>true</code> if the underlying method <em>might</em>
    *          accept a var-arg list; <code>false</code> if it definitely
    *          cannot. If <code>true</code>, the fact whether the method
    *          accepts var-args will be determined for certain in the
    *          constructor. Only methods of registered callback libraries
    *          can support var-arg lists.
    *
    * @throws  ClassCastException
    *          if <code>method</code> cannot be invoked on an instance of
    *          <code>target</code>; if <code>method</code>'s return type is
    *          not (even leniently) {@link #isAssignmentCompatible assignment
    *          compatible} with <code>returnType</code>.
    */
   Function(Class target,
            Class returnType,
            Method method,
            Class[] parameterTypes,
            boolean enableVarArgs)
   {
      if (!method.getDeclaringClass().isAssignableFrom(target))
      {
         throw new ClassCastException(
            "Method '" + method + "' cannot be invoked on an instance of " +
            target);
      }
      
      if (!isAssignmentCompatible(method.getReturnType(), returnType, false))
      {
         throw new ClassCastException(
            "Method '" + method + "' returns " + method.getReturnType() +
            ", which is not compatible with specified return type " +
            returnType);
      }
      
      this.target = target;
      this.returnType = returnType;
      this.method = method;
      
      // Test for a var-args type signature, where the last argument is
      // an array of objects. Only relevant if var-args support is enabled.
      Class[] sig = method.getParameterTypes();
      int len = sig.length;
      Class lastParm = null;
      if (enableVarArgs                     &&
          len > 0                           &&
          (lastParm = sig[--len]).isArray() &&
          !lastParm.getComponentType().isPrimitive())
      {
         minArgs = len;
      }
      
      if (parameterTypes != null)
      {
         this.parameterTypes = parameterTypes;
      }
      else
      {
         this.parameterTypes = method.getParameterTypes();
      }
   }
   
   /**
    * Compose a string indicating a full function signature. Uses the form:
    * <pre>
    *    returnType name([parm1[, parm2[..., parmN]]])
    * </pre>
    *
    * @param   name
    *          Function name.
    * @param   returnType
    *          Type which this function returns.
    * @param   minArgs
    *          Minimum number of required arguments. If <code>-1</code>,
    *          var-args support is disabled.
    * @param   signature
    *          Array of classes representing the parameters this function
    *          expects, in order.
    *
    * @return  String in the form expressed above.
    */
   public static String asString(String name,
                                 Class returnType,
                                 int minArgs,
                                 Class[] signature)
   {
      StringBuffer buf = new StringBuffer();
      
      buf.append(returnType.getName())
         .append(' ')
         .append(name)
         .append('(');
      
      // Loop through parameters and append their class names.
      if (signature != null)
      {
         int len = (minArgs >= 0 ? minArgs : signature.length);
         for (int i = 0; i < len; i++)
         {
            if (i > 0)
            {
               buf.append(", ");
            }
            Class next = signature[i];
            buf.append(next == null ? "null" : next.getName());
         }
      }
      
      // Tack on elipsis to represent var-args.
      if (minArgs >= 0)
      {
         buf.append(", ...");
      }
      
      buf.append(')');
      
      return buf.toString();
   }
   
   /**
    * Compare the specified argument signature against the specified match
    * signature to determine whether <code>match</code> is compatible with
    * <code>sig</code>. They match if they contain the same number of
    * minimum required parameters and for each index <cod>i</code>, either:
    * <ul>
    * <li><code>match[i]</code> is compatible with <code>sig[i]</code>; or
    * <li><code>match[i] == null</code>.
    * </ul>
    * Types are tested for assignment compatibility rather than equality.
    * Assignment compatibility may be <em>strict</em> or <em>lenient</em>:
    * <ul>
    * <li>A "strict" signature match: each required parameter type must be
    *     assignable from the provided parameter type, OR they must be the
    *     same primitive types or at least represent a compatible
    *     primitive/wrapper pair (e.g., <code>int</code> and
    *     <code>java.lang.Integer</code>).
    * <li>A "lenient" signature match: each required parameter type must be
    *     assignable from the provided parameter type, OR they must be
    *     compatible boolean types (primitive/wrapper combination OK), OR
    *     they must be any combination of numeric primitive and/or wrapper
    *     types.
    * </ul>
    *
    * @param   sig
    *          Array of argument types which define the baseline signature.
    * @param   match
    *          Array of classes which indicate a parameter signature we are
    *          trying to match; <code>null</code>s are permitted.
    * @param   minArgs
    *          Minimum arguments required by the baseline signature. If
    *          negative, it is assumed var-arg lists are not to be
    *          considered.
    * @param   strict
    *          Whether to enforce a strict match or not.
    *
    * @return  <code>true</code> if the signatures match as specified above,
    *          else <code>false</code>.
    */
   public static boolean matchSignature(Class[] sig,
                                        Class[] match,
                                        int minArgs,
                                        boolean strict)
   {
      int len1 = (sig == null ? 0 : sig.length);
      int len2 = (match == null ? 0 : match.length);
      
      // If supporting var-args, last parameter is the variable argument
      // object array, so we don't want to check the last parameter in the
      // loop below.
      if (minArgs >= 0)
      {
         len1--;
         
         // Check for too few parameters.
         if (len2 < minArgs)
         {
            return false;
         }
      }
      else if (len1 != len2)
      {
         // For non var-arg functions, a mismatched number of args is right
         // out.
         return false;
      }
      
      // Compare each parameter in this func with the parameter at the same
      // position in target. A null parameter in target always matches.
      // Otherwise, parameters must be compatible, as defined by type map.
      // For functions which support variable argument lists, we must match
      // only the required arguments. Otherwise, we must match all the
      // arguments.
      for (int i = 0; i < len1; i++)
      {
         Class next = match[i];
         if (next == null)
         {
            next = UnknownType.class;
         }
         
         if (!isAssignmentCompatible(sig[i], next, strict))
         {
            return false;
         }
      }
      
      return true;
   }
   
   /**
    * Test a specific pair of data types for assignment compatibility.
    * Assignment compatibility may be <em>strict</em> or <em>lenient</em>:
    * <ul>
    * <li>A "strict" match: the required parameter type must be assignable
    *     from the provided parameter type, OR they must be the same
    *     primitive types or at least represent a compatible primitive/wrapper
    *     pair (e.g., <code>int</code> and <code>java.lang.Integer</code>).
    * <li>A "lenient" match: each required parameter type must be assignable
    *     from the provided parameter type, OR they must be compatible
    *     boolean types (primitive/wrapper combination OK), OR they must be
    *     any combination of numeric primitive and/or wrapper types.
    * </ul>
    *
    * @param   required
    *          Required data type.
    * @param   provided
    *          Provided data type.
    * @param   strict
    *          Whether to perform a strict match, as defined above.
    *
    * @return  <code>true</code> if a value of type <code>provided</code>
    *          can be assigned (possibly with some conversion required) to
    *          a parameter/variable of type <code>required</code>, given
    *          the strictness constraint specified. Otherwise,
    *          <code>false</code>.
    */
   public static boolean isAssignmentCompatible(Class required,
                                                Class provided,
                                                boolean strict)
   {
      // Void type only matches itself.
      if (required == Void.TYPE || provided == Void.TYPE)
      {
         return (required == provided);
      }
      
      // Unknown type matches everything (non-void).
      if (provided == UnknownType.class)
      {
         return true;
      }
                         
      boolean reqPrim = required.isPrimitive();
      boolean prvPrim = provided.isPrimitive();
      
      // Get the easy cases out of the way first...
      if (!reqPrim && !prvPrim)
      {
         return required.isAssignableFrom(provided);
      }
      
      if (strict)
      {
         int len = COMPAT_PAIRS.length;
         for (int i = 0; i < len; i += 2)
         {
            Class prim = COMPAT_PAIRS[i];
            Class ref = COMPAT_PAIRS[i + 1];
            if ((required == prim || required == ref) &&
                (provided == prim || provided == ref))
            {
               return true;
            }
         }
         
         return false;
      }
      
      if (required == Boolean.TYPE || required == Boolean.class)
      {
         return (provided == Boolean.TYPE || provided == Boolean.class);
      }
      
      if (reqPrim || Number.class.isAssignableFrom(required))
      {
         // With enough casting, any numeric conversion is possible. We may
         // lose precision, but that is up to the expression to detect at
         // runtime, if the user chooses to enable the check.
         return (provided == Long.TYPE      || provided == Long.class      ||
                 provided == Integer.TYPE   || provided == Integer.class   ||
                 provided == Double.TYPE    || provided == Double.class    ||
                 provided == Float.TYPE     || provided == Float.class     ||
                 provided == Short.TYPE     || provided == Short.class     ||
                 provided == Character.TYPE || provided == Character.class ||
                 provided == Byte.TYPE      || provided == Byte.class);
      }
      
      if (required == Object.class)
      {
         return true;
      }
      
      return false;
   }
   
   /**
    * Compose a string indicating a full function signature. Use the form:
    * <pre>
    *    returnType name([parm1[, parm2[..., parmN]]])
    * </pre>
    *
    * @return  String in the form expressed above.
    */
   public String toString()
   {
      String sig = asString(method.getName(),
                            getReturnType(),
                            minArgs,
                            method.getParameterTypes());
      
      StringBuffer buf = new StringBuffer(sig);
      buf.append(" @")
         .append(getTarget());
      
      if (prerequisite != null)
      {
         buf.append(" <---")
            .append(prerequisite);
      }
      
      return buf.toString();
   }
   
   /**
    * Get the type of method this function object represents: either one of
    * the bootstrap types or a general purpose method.
    *
    * @return  A constant indicating the underlying method's type.
    *
    * @see     #GENERAL
    * @see     #GET_LIB
    * @see     #GET_VAR
    * @see     #SET_VAR
    */
   int getType()
   {
      return type;
   }
   
   /**
    * Set the type of method this function object represents: either one of
    * the bootstrap types or a general purpose method.
    *
    * @param   type
    *          A constant indicating the underlying method's type.
    *
    * @see     #GENERAL
    * @see     #GET_LIB
    * @see     #GET_VAR
    * @see     #SET_VAR
    */
   void setType(int type)
   {
      this.type = type;
   }
   
   /**
    * Get the optional symbol which represents a name or identifier for the
    * invocation target object. This symbol is set by the symbol resolver.
    * It is used when parsing an expressions to set the text of an AST node
    * which represents a bootstrap method. Since these methods are implicitly
    * added to the AST by the symbol resolver, there is no text in the
    * expression which is naturally available for this purpose.
    *
    * @return  A symbol to be associated with a bootstrap method AST node
    *          in the postfix form of a parsed expression.
    */
   String getSymbol()
   {
      return symbol;
   }
   
   /**
    * Set the optional symbol which represents a name or identifier for the
    * invocation target object. This symbol is set by the symbol resolver.
    * It is used when parsing an expressions to set the text of an AST node
    * which represents a bootstrap method. Since these methods are implicitly
    * added to the AST by the symbol resolver, there is no text in the
    * expression which is naturally available for this purpose.
    *
    * @param   symbol
    *          A symbol to be associated with a bootstrap method AST node
    *          in the postfix form of a parsed expression.
    */
   void setSymbol(String symbol)
   {
      this.symbol = symbol;
   }
   
   /**
    * Get a <code>Function</code> object which encapsulates another method
    * associated with this object. The returned object represents a
    * <a href=#bootstrap>bootstrap method</a> which must be invoked as a
    * prerequisite to invoking the method represented by <em>this</em>
    * <code>Function</code> instance. Invocation of the bootstrap method
    * load's the target object upon which the method represented by this
    * object is invoked.
    *
    * @return  A <code>Function</code> which describes this object's
    *          prerequisite bootstrap method. If no such prerequisite exists,
    *          <code>null</code> is returned.
    *
    * @see     #setPrerequisite
    */
   Function getPrerequisite()
   {
      return prerequisite;
   }
   
   /**
    * Set the <code>Function</code> instance which represents the
    * prerequisite <a href=#bootstrap>bootstrap method</a> for this object.
    * This method should only be called by the {@link SymbolResolver}.
    *
    * @param   prerequisite
    *          A <code>Function</code> which describes this object's
    *          prerequisite bootstrap method.
    *
    * @see     #getPrerequisite
    *
    * @throws  ClassCastException
    *          if the object returned by the specified bootstrap method
    *          does not match the target object type required by this
    *          function.
    */
   void setPrerequisite(Function prerequisite)
   {
      if (!getTarget().isAssignableFrom(prerequisite.getReturnType()))
      {
         throw new ClassCastException(
            "Prerequisite function (" + prerequisite + ") returns " +
            prerequisite.getReturnType() +
            ", which is not compatible with this function's target class (" +
            getTarget() + ")");
      }
      
      this.prerequisite = prerequisite;
   }
   
   /**
    * Get the invocation target class of this function.
    *
    * @return  Invocation target class.
    */
   Class getTarget()
   {
      return target;
   }
   
   /**
    * Get the method backing this function.
    *
    * @return  Method.
    */
   Method getMethod()
   {
      return method;
   }
   
   /**
    * Get the return type of this function, which may be more specific than
    * the return type of the underlying method.
    *
    * @return  Return type.
    */
   Class getReturnType()
   {
      return returnType;
   }
   
   /**
    * Get the minimum number of arguments required by this function. This
    * property is only relevant if this function supports var-args.
    *
    * @return  Minimum number of required arguments, or <code>-1</code> if
    *          this function does not support variable argument lists.
    */
   int getMinArgs()
   {
      return minArgs;
   }
   
   /**
    * Get the array of parameter types expected by this function object.
    * This will be different than the parameter types expected by the
    * underlying method if this function represents a callback library
    * method which accepts a variable-length argument list. In that case,
    * the underlying method will expect an array of type <code>Object</code>
    * as its last parameter, while this method will break out the specific
    * types of each of the variable-length list arguments.
    *
    * @return  Parameter types expected by this function.
    */
   Class[] getParameterTypes()
   {
      return parameterTypes;
   }
   
   /**
    * Compare the specified argument signature against this function's
    * signature to determine whether <code>match</code> is compatible with
    * this function's signature. They match if they contain the same number
    * of minimum required parameters and for each index <cod>i</code>, either:
    * <ul>
    * <li><code>match[i]</code> is compatible with this function's parameter
    *     at the same position; or
    * <li><code>match[i] == null</code>.
    * </ul>
    * Types are tested for assignment compatibility rather than equality.
    *
    * @param   match
    *          Array of classes which indicate a parameter signature we are
    *          trying to match; <code>null</code>s are permitted.
    *
    * @return  <code>true</code> if the signatures match as specified above,
    *          else <code>false</code>.
    */
   boolean matchSignature(Class[] match, boolean strict)
   {
      return matchSignature(method.getParameterTypes(),
                            match,
                            minArgs,
                            strict);
   }
}
